import copy
from dataclasses import asdict, dataclass
import datetime
from typing import Any, Dict, Optional
from uuid import UUID, uuid4


class BasicParams:
    """This class is used to create the basic parameters for the API calls. 
        These parameters are common to all the API calls.
    """
    
    def __init__(
        self,
        lookup_date: datetime.date | None = None,
        dateformat: str = "timestamp",
        human: bool = False,
        offset: int | None = None,
        limit: int = 500,
        format: str = "json",
        data_headers: bool = True,
    ):
        """Initializes the BasicParams class with the basic parameters for the API calls.

        Args:
            lookup_date (date, optional): Specific date to lookup. Defaults to None.
            dateformat (str, optional): The dateformat parameter allows you specify the format you wish to receive date and time information in. "timestamp", "spreadsheet", "unix". Defaults to "timestamp".
            human (bool, optional): se human-readable attribute names in the JSON or CSV output instead of the standard camelCase attribute names. Defaults to False.
            offset (int, optional): Used with limit to allow you to implement pagination in your application. Offset returns values starting at a certain value. Defaults to None.
            limit (int, optional): Limit the number of results for a particular API call or override an endpoint's default limits to get more data. Defaults to 500.
            format (str, optional): The format parameter allows you to specify the format you wish to receive the data in. "json", "csv". Defaults to "json".
            data_headers (bool, optional): Used to turn off headers when using CSV output. Defaults to None.
        """
        self.params = {}
        if lookup_date:
            self.params['date'] = lookup_date.strftime('%Y-%m-%d')
        if dateformat:
            self.params['dateformat'] = dateformat
        if human:
            self.params['human'] = human
        if offset:
            self.params['offset'] = offset
        if limit:
            self.params['limit'] = limit
        if format:
            self.params['format'] = format
        if data_headers:
            self.params['headers'] = data_headers

    def __str__(self) -> str:
        return str(self.params)
    
    def to_dict(self) -> Dict[str, Any]:
        return self.params
                
class FromToParams:
    """
    This class is used to create the from and to parameters for the API calls.
    These are used with API calls that can select a range of dates.
    """
    
    def __init__(
        self,
        from_date: datetime.date | None = None,
        to_date: datetime.date | None = None,
        countback: int | None = None,
    ):
        """Initializes the FromToParams class with the from and to dates and countback.

        Args:
            from_date (date, optional): Limit the status to dates after from (inclusive). Should be combined with to to create a range. Defaults to None.
            to_date (date, optional): Limit the status to dates before to (inclusive). Should be combined with from to create a range. Defaults to None.
            countback (int, optional): Countback will fetch a number of dates before (to the left of) to. If you use from, countback is not required. Defaults to None.
        """
        self.params = {}
        if from_date:
            self.params['from'] = from_date.strftime('%Y-%m-%d')
        if to_date:
            self.params['to'] = to_date.strftime('%Y-%m-%d')
        if countback:
            self.params['countback'] = countback
            
    def __str__(self) -> str:
        return str(self.params)
    
    def to_dict(self) -> Dict[str, Any]:
        return self.params
    
@dataclass
class OptionsChainParams:
    underlying: str
    basic_params: Optional[BasicParams] = None
    from_to_params: Optional[FromToParams] = None
    expiration: Optional[datetime.date] = None
    month: Optional[int] = None
    year: Optional[int] = None
    weekly: Optional[bool] = None
    monthly: Optional[bool] = None
    quarterly: Optional[bool] = None
    dte: Optional[int] = None    
    feed: Optional[str] = None  # "live" or "cached"        
    side: str = "both"
    range: str = "all"
    strike: Optional[float] = None
    strikeLimit: Optional[int] = None
    minOpenInterest: Optional[int] = None
    minVolume: Optional[int] = None
    maxBidAskSpread: Optional[float] = None
    maxBidAskSpreadPct: Optional[float] = None
    nonstandard: bool = False
    columns: Optional[str] = None
    output: str = "dataframe"
    id: str = str(uuid4())

    def to_dict(self):
        def serialize(v):
            try:
                return v.to_dict()
            except AttributeError:
                if isinstance(v, (datetime.date, UUID)):
                    return str(v)
                return v

        out = {k: serialize(v) for k, v in asdict(self).items() if v is not None}
        del out['id']
        return out

    def make_copy(self, **kwargs) -> 'OptionsChainParams':
        new_params = copy.deepcopy(self)
        
        # Update the new object with the provided overrides
        for key, value in kwargs.items():
            if hasattr(new_params, key):
                setattr(new_params, key, value)
            else:
                raise ValueError(f"Invalid parameter: {key}")
            
        # Create a new UUID for the new object
        new_params.id = uuid4()
                
        return new_params

@dataclass
class OptionsQuoteParams:
    option_symbol: str
    basic_params: Optional[BasicParams] = None
    from_to_params: Optional[FromToParams] = None
    columns: Optional[str] = None
    output: str = "dataframe"

    def to_dict(self):
        out = {k: v for k, v in asdict(self).items() if v is not None}
        return out

    def make_copy(self, **kwargs) -> 'OptionsQuoteParams':
        new_params = copy.deepcopy(self)
        
        # Update the new object with the provided overrides
        for key, value in kwargs.items():
            if hasattr(new_params, key):
                setattr(new_params, key, value)
            else:
                raise ValueError(f"Invalid parameter: {key}")
            
        return new_params


import os
from typing import Optional

def get_api_key() -> Optional[str]:
    """Get API key from environment variable or raise helpful error."""
    api_key = os.environ.get('MARKET_DATA_API_KEY')
    if not api_key:
        raise ValueError(
            "MARKET_DATA_API_KEY environment variable not set. "
            "Please set it with your API key from marketdata.app"
        )
    return api_key

MARKET_DATA_API_KEY = get_api_key()


"""
marketdata.app API Client
"""
import os
from time import time
from urllib.parse import urlencode
import requests
import datetime
import json
import hashlib
 
import numpy as np
import pandas as pd
from loguru import logger
import pandas as pd

from marketdata.client_params import BasicParams, FromToParams, OptionsChainParams
from marketdata.credentials import MARKET_DATA_API_KEY

BASE_URL = 'https://api.marketdata.app/v1/'

class MarketDataClient:
    
    def __init__(self) -> None:
        self.BASE_URL = 'https://api.marketdata.app/v1/'
        self.api_key = MARKET_DATA_API_KEY
        self.headers = {
            'Content-Type': 'application/json',
            'Authorization': f'Bearer {self.api_key}'
        }
        self.api_calls = 0
    
    def handle_response(self, response, output):
        try:
            if output == "raw":
                return response.json(), response.status_code
            elif output == "dataframe":
                if 200 <= response.status_code < 300:
                    df = pd.DataFrame(response.json())
                    try:
                        df.drop(columns=['s'], inplace=True)
                    except KeyError:
                        pass
                    return df, response.status_code
                else:
                    return response.json(), response.status_code                
        # Handle the case where the response is not JSON
        except ValueError:
            return response.text, response.status_code
    
    # /v1/funds/candles/{resolution}/{symbol}/
    def get_fund_candles(
        self,
        resolution: str,
        symbol: str,
        basic_params: BasicParams | None = None,
        from_to_params: FromToParams | None = None,
        columns: str = None,
        output = "dataframe"
    ):
        url = BASE_URL + f'funds/candles/{resolution}/{symbol}/'
        params = {}
        
        if basic_params:
            params.update(basic_params.params)
        if from_to_params:
            params.update(from_to_params.params)
        if columns:
            params['columns'] = columns
        
        response = requests.get(url, headers=self.headers, params=params)
        self.api_calls += 1
        return self.handle_response(response, output)

    # /v1/indices/candles/{resolution}/{symbol}/
    def get_index_candles(
        self,
        resolution: str,
        symbol: str,
        basic_params: BasicParams | None = None,
        from_to_params: FromToParams | None = None,
        columns: str = None,
        output = "dataframe"
    ):
        url = BASE_URL + f'indices/candles/{resolution}/{symbol}/'
        params = {}
        
        if basic_params:
            params.update(basic_params.params)
        if from_to_params:
            params.update(from_to_params.params)
        if columns:
            params['columns'] = columns
        
        response = requests.get(url, headers=self.headers, params=params)
        self.api_calls += 1
        return self.handle_response(response, output)

    # /v1/indicies/quotes/{symbol}/
    def get_index_quote(
        self,
        symbol: str,
        basic_params: BasicParams | None = None,
        columns: str = None,
        output = "dataframe"
    ):
        url = BASE_URL + f'indices/quotes/{symbol}/'
        params = {}
        
        if basic_params:
            params.update(basic_params.params)
        if columns:
            params['columns'] = columns
        
        # Remove the lookup_date parameter if it exists
        if isinstance(basic_params, dict) and basic_params.get('lookup_date'):
            del basic_params['lookup_date']
            logger.warning("The lookup_date parameter is not supported for this endpoint. It will be ignored.")
        
        response = requests.get(url, headers=self.headers, params=params)
        self.api_calls += 1
        return self.handle_response(response, output)
            
    # /v1/markets/status/
    def get_markets_status(
        self,
        basic_params: BasicParams | None = None,
        from_to_params: FromToParams | None = None,
        country: str = "US",
        output = "dataframe"
    ) -> dict | str:
        """Get market status ("open" or "closed") for a date or range of dates.

        Args:
            
            basic_params (BasicParams, optional): See BasicParams class. Defaults to None.
            from_to_params (FromToParams, optional): See FromToParams class. Defaults to None.
            country (str, optional): Use to specify the country of the exchange. Use the two digit ISO 3166 country code. Defaults to "US".
            output (str, optional): The output format. Can be "dataframe", which is a pandas dataframe or "raw", which is the raw JSON response. Defaults to "dataframe".
            
        Returns:
            dict, str: Returns the raw data and the status code.
        """
        url = BASE_URL + 'markets/status/'
        
        params = {}
        if basic_params:
            params.update(basic_params.params)
        if from_to_params:
            params.update(from_to_params.params)
        params['country'] = country
        
        response = requests.get(url, headers=self.headers, params=params)
        self.api_calls += 1
        return self.handle_response(response, output)
        
    # /v1/options/chain/{underlying}/
    def get_options_chain(self, params: OptionsChainParams):
        """Get options chain for a symbol.

        Args:
            params (OptionsChainParams): An object containing all the parameters for the options chain request.

        Returns:
            dict, str: Returns the raw data as a dictionary or an error message.
        """
        url = BASE_URL + f'options/chain/{params.underlying}/'
        request_params = params.to_dict()

        # Remove 'underlying' and 'id' from request_params as they're not needed in the API call
        request_params.pop('underlying', None)
        request_params.pop('id', None)

        # Handle nested objects
        if params.basic_params:
            request_params.update(params.basic_params.params)
            request_params.pop('basic_params', None)
        if params.from_to_params:
            request_params.update(params.from_to_params.params)
            request_params.pop('from_to_params', None)

        logger.warning(f"Debug - Full URL: {url}?{urlencode(request_params)}")
        url = f"{url}?{urlencode(request_params)}"
        response = requests.get(url, headers=self.headers)
        self.api_calls += 1
        return self.handle_response(response, params.output)

    # /v1/options/expirations/{underlying}/
    def get_options_expirations(
        self,
        underlying: str,
        basic_params: BasicParams | None = None,
        strike: float | None = None,
        columns: str = None,
        output = "dataframe"
    ):
        url = BASE_URL + f'options/expirations/{underlying}/'
        params = {}
        if basic_params:
            params.update(basic_params.params)
        if strike:
            params['strike'] = strike
        if columns:
            params['columns'] = columns
        
        response = requests.get(url, headers=self.headers, params=params)
        self.api_calls += 1
        return self.handle_response(response, output)
    

    # /v1/options/lookup/{userInput}/
    # This doesn't work in the Swagger UI, so not implemented here
        

    # /v1/options/quotes/{optionSymbol}/
    def get_options_quotes(
        self,
        option_symbol: str,
        basic_params: BasicParams | None = None,
        from_to_params: FromToParams | None = None,
        columns: str = None,
        output = "dataframe"
    ):
        url = BASE_URL + f'options/quotes/{option_symbol}/'
        params = {}
        if basic_params:
            params.update(basic_params.params)
        if from_to_params:
            params.update(from_to_params.params)
        if columns:
            params['columns'] = columns

        response = requests.get(url, headers=self.headers, params=params)
        self.api_calls += 1
        return self.handle_response(response, output)
        

    # /v1/options/strikes/{underlying}/
    def get_options_strikes(
        self,
        underlying: str,
        basic_params: BasicParams | None = None,
        expiration_date: datetime.date | None = None,
        columns: str = None,
        output: str = "dataframe"
    ):
        """Get a list of current or historical options strikes for an underlying
        symbol. If no optional parameters are used, the endpoint returns the 
        strikes for every expiration in the chain.

        Args:
            underlying (str): The underlying ticker symbol for the options chain you wish to lookup. Ticker Formats: (TICKER, TICKER.EX, EXCHANGE:TICKER)
            basic_params (BasicParams | None, optional): See BasicParams class. Defaults to None.
            expiration_date (datetime.date | None, optional): Limit the strikes to a specific expiration date. Defaults to None.
            columns (str, optional): Limits the results and only request the columns you need. The most common use of this feature is to embed a single numeric result from one of the end points in a spreadsheet cell.. Defaults to None.
            output (str, optional): The output format. Can be "dataframe", which is a pandas dataframe or "raw", which is the raw JSON response. Defaults to "dataframe".

        Returns:
            _type_: _description_
        """
        url = BASE_URL + f'options/strikes/{underlying}/'
        params = {}
        if basic_params:
            params.update(basic_params.params)
        if expiration_date:
            params['expiration'] = expiration_date.strftime('%Y-%m-%d')
        if columns:
            params['columns'] = columns
            
        response = requests.get(url, headers=self.headers, params=params)
        self.api_calls += 1
        return self.handle_response(response, output)
    

    # /v1/stocks/bulkcandles/{resolution}/
    def get_bulk_stock_candles(
        self,
        symbols: list[str],
        basic_params: BasicParams | None = None,
        from_to_params: FromToParams | None = None,
        exchange: str = None,
        country: str = "US",
        snapshot: bool = False,
        adjust_splits: bool = None,
        adjust_dividends: bool = None,
        columns: str = None,
        output = "dataframe"
    ):
        """This seems to be incorrectly implemented in the API. Can't get
        it to work and the Swagger UI gives invalid input fields.
        """
        url = BASE_URL + f'stocks/bulkcandles/daily/'
        params = {}
        if basic_params:
            params.update(basic_params.params)
        if from_to_params:
            params.update(from_to_params.params)
        if exchange:
            params['exchange'] = exchange
        if adjust_splits:
            params['adjust_splits'] = adjust_splits
        if adjust_dividends:
            params['adjust_dividends'] = adjust_dividends
        if columns:
            params['columns'] = columns
        params['symbols'] = ','.join(symbols)
        params['country'] = country
        params['snapshot'] = snapshot
        
        response = requests.get(url, headers=self.headers, params=params)
        self.api_calls += 1
        return self.handle_response(response, output)

    # /v1/stocks/bulkquotes/
    def get_bulk_stock_quotes(
        self,
        symbols: list[str],
        basic_params: BasicParams | None = None,
        columns: str = None,
        output = "dataframe"
    ):
        url = BASE_URL + 'stocks/bulkquotes/'
        params = {}
        if basic_params:
            params.update(basic_params.params)
        if columns:
            params['columns'] = columns
        params['symbols'] = ','.join(symbols)
        
        # If the lookup_date is in the params, remove it
        if isinstance(basic_params, dict) and basic_params.get('lookup_date'):
            del basic_params['lookup_date']
            logger.warning("The lookup_date parameter is not supported for this endpoint. It will be ignored.")
        
        response = requests.get(url, headers=self.headers, params=params)
        # TODO: Verify that this is actually counted as a single API call
        self.api_calls += 1
        return self.handle_response(response, output)

    # /v1/stocks/candles/{resolution}/{symbol}/
    def get_stock_candles(
        self,
        symbol: str,
        resolution: str = "1D",
        basic_params: BasicParams | None = None,
        from_to_params: FromToParams | None = None,
        columns: str = None,
        exchange: str = None,
        extended_hours: bool = False,
        exchange_country: str = "US",
        adjust_splits: bool = None,
        adjust_dividends: bool = None,
        output = "dataframe"
    ):
        """Get historical stock candles for a symbol.
        
        Args:
            symbol (str): The company's ticker symbol. If no exchange is specified, by default a US exchange will be assumed. You may embed the exchange in the ticker symbol using the Yahoo Finance or TradingView formats. Ticker Formats: (TICKER, TICKER.EX, EXCHANGE:TICKER)
            resolution (str, optional): The duration of each candle. Minutely Resolutions: (1, 3, 5, 15, 30, 45, ...) Hourly Resolutions: (H, 1H, 2H, ...) Daily Resolutions: (D, 1D, 2D, ...) Weekly Resolutions: (W, 1W, 2W, ...) Monthly Resolutions: (M, 1M, 2M, ...) Yearly Resolutions:(Y, 1Y, 2Y, ...)'). Defaults to "1D".
            basic_params (BasicParams, optional): See BasicParams class. Defaults to None.
            from_to_params (FromToParams, optional): See FromToParams class. Defaults to None.
            columns (str, optional): Limits the results and only request the columns you need. The most common use of this feature is to embed a single numeric result from one of the end points in a spreadsheet cell.. Defaults to None.
            exchange (str, optional): Use to specify the exchange of the ticker. This is useful when you need to specify stock that quotes on several exchanges with the same symbol. You may specify the exchange using the EXCHANGE ACRONYM, MIC CODE, or two digit YAHOO FINANCE EXCHANGE CODE. If no exchange is specified symbols will be matched to US exchanges first.. Defaults to None.
            extended_hours (bool, optional): Include extended hours trading sessions when returning intraday candles. Daily resolutions never return extended hours candles. Defaults to False.
            exchange_country (str, optional): Specify the country of the exchange (not the country of the company) in conjunction with the symbol argument. This argument is useful when you know the ticker symbol and the country of the exchange, but not the exchange code. Use the two digit ISO 3166 country code. Defaults to "US".
            adjust_splits (bool, optional): Adjust historical data for for historical splits and reverse splits. Market Data uses the CRSP methodology for adjustment. Daily candles default: true. Intraday candles default: false.
            adjust_dividends (bool, optional): Adjust candles for dividends. Market Data uses the CRSP methodology for adjustment. Daily candles default: true. Intraday candles default: false.
            output (str, optional): The output format. Can be "dataframe", which is a pandas dataframe or "raw", which is the raw JSON response. Defaults to "dataframe".
        Returns:
            _type_: _description_
        """
        
        url = BASE_URL + f'stocks/candles/{resolution}/{symbol}/'
        params = {}
        if basic_params:
            params.update(basic_params.params)
        if from_to_params:
            params.update(from_to_params.params)
            
        options = {
            'columns': columns,
            'exchange': exchange,
            'extended_hours': extended_hours,
            'exchange_country': exchange_country,
            'adjust_splits': adjust_splits,
            'adjust_dividends': adjust_dividends
        }
        
        # Update params with non-None options
        for key, value in options.items():
            if value is not None:
                params[key] = value

        response = requests.get(url, headers=self.headers, params=params)
        self.api_calls += 1
        return self.handle_response(response, output)
    
    # /v1/stocks/earnings/{symbol}/
    def get_earnings(
            self,
            symbol: str,
            basic_params: BasicParams | None = None,
            from_to_params: FromToParams | None = None,
            report: str = None,
            columns: str = None,
            output: str = "dataframe"
        ):
        url = BASE_URL + f'stocks/earnings/{symbol}/'
        params = {}

        # Unpack parameters from objects if they are not None
        if basic_params:
            params.update(basic_params.params)
        if from_to_params:
            params.update(from_to_params.params)

        # Update params with other optional parameters if they are not None
        optional_params = {'report': report, 'columns': columns}
        params.update({k: v for k, v in optional_params.items() if v is not None})
        
        response = requests.get(url, headers=self.headers, params=params)
        self.api_calls += 1
        return self.handle_response(response, output)

    # /v1/stocks/news/{symbol}/
    def get_stock_news(
        self,
        symbol: str,
        basic_params: BasicParams = None,
        from_to_params: FromToParams = None,
        output = "dataframe"
    ):
        url = BASE_URL + f'stocks/news/{symbol}/'
        params = {}
        
        if basic_params:
            params.update(basic_params.params)
        if from_to_params:
            params.update(from_to_params.params)
        
        response = requests.get(url, headers=self.headers, params=params)
        self.api_calls += 1
        return self.handle_response(response, output)

    # /v1/stocks/quotes/{symbol}/
    def get_stock_quote(
        self,
        symbol: str,
        basic_params: BasicParams = None,
        columns: str = None,
        output = "dataframe"
    ):
        url = BASE_URL + f'stocks/quotes/{symbol}/'
        params = {}
        
        if basic_params:
            params.update(basic_params.params)
        if columns:
            params['columns'] = columns
        
        # This is one of the few API calls that does not have a lookup_date,
        # so we need to remove it if it exists
        if isinstance(basic_params, dict) and basic_params.get('lookup_date'):
            del basic_params['lookup_date']
        
        response = requests.get(url, headers=self.headers, params=params)
        self.api_calls += 1
        return self.handle_response(response, output)
    


"""
marketdata.app API Client
"""
import os
import sys
from time import time
from urllib.parse import urlencode
import datetime


import pandas as pd
from loguru import logger
import pandas as pd

import aiohttp
from aiohttp import TCPConnector, ClientSession
import asyncio
from typing import Any, Dict, List, Optional, Tuple
import datetime

from marketdata import BasicParams, FromToParams
from marketdata import MARKET_DATA_API_KEY
from marketdata.client_params import OptionsChainParams, OptionsQuoteParams

BASE_URL = 'https://api.marketdata.app/v1/'

class MarketDataAsyncClient:
    
    def __init__(self) -> None:
        self.BASE_URL = 'https://api.marketdata.app/v1/'
        self.api_key = MARKET_DATA_API_KEY
        self.headers = {
            'Content-Type': 'application/json',
            'Authorization': f'Bearer {self.api_key}'
        }
        self.api_calls = 0

    async def handle_response_async(self, response, output):
        try:
            if output == "raw":
                try:
                    return await response.json(), response.status
                except aiohttp.client_exceptions.ContentTypeError as e:
                    logger.error(f"Error getting response: {e}")
                    return await response.text(), response.status
            elif output == "dataframe":
                if 200 <= response.status < 300:
                    json_data = await response.json()
                    df = pd.DataFrame(json_data)
                    try:
                        df.drop(columns=['s'], inplace=True)
                    except KeyError:
                        pass
                    return df, response.status
                else:
                    try:
                        return await response.json(), response.status
                    except Exception as e:
                        logger.error(f"Error getting response: {e}")
                        logger.error(await response.text())
                        exit()
        # Handle the case where the response is not JSON
        except ValueError:
            return await response.text(), response.status
        
        
    # /v1/options/chain/{underlying}/
    async def get_options_chain(self, params: OptionsChainParams):
        """Get options chain for a symbol asynchronously."""

        url = BASE_URL + f'options/chain/{params.underlying}/'
        api_params = params.to_dict()
        if params.basic_params:
            api_params.update(params.basic_params.params)
        if params.from_to_params:
            api_params.update(params.from_to_params.params)

        async with aiohttp.ClientSession() as session:
            for key, value in api_params.items():
                if isinstance(value, bool):
                    api_params[key] = str(value).lower()
                elif isinstance(value, datetime.date):
                    api_params[key] = value.strftime('%Y-%m-%d')
            
            if 'basic_params' in api_params:
                if 'dateformat' in api_params['basic_params']:
                    api_params['dateformat'] = api_params['basic_params']['dateformat']
                if 'date' in api_params['basic_params']:
                    api_params['date'] = api_params['basic_params']['date']
                if 'format' in api_params['basic_params']:
                    api_params['format'] = api_params['basic_params']['format']
                if 'headers' in api_params['basic_params']:
                    api_params['headers'] = api_params['basic_params']['headers']
                if 'limit' in api_params['basic_params']:
                    api_params['limit'] = api_params['basic_params']['limit']
                del api_params['basic_params']
                
            url = f"{url}?{urlencode(api_params)}"
            async with session.get(url, headers=self.headers) as response:
                url_with_token = url + '&token=' + self.api_key
                # logger.debug(f"API call (Option Chain): {url_with_token}")
                self.api_calls += 1
                data, status_code = await self.handle_response_async(response, params.output)
                if not 200 <= status_code < 300:
                    # logger.error(f"Error fetching options chain for {params.underlying}. Status code: {status_code}")
                    pass

        if isinstance(data, dict) and 's' in data and data.get('s') == 'error':
            logger.error(data)
            return data
        else:
            return data
        

    # /v1/options/quotes/{optionSymbol}/
    async def get_options_quotes(self, params: OptionsQuoteParams):

        url = BASE_URL + f'options/quotes/{params.option_symbol}/'
        api_params = {}
        if params.basic_params:
            api_params.update(params.basic_params.params)
        if params.from_to_params:
            api_params.update(params.from_to_params.params)
        if params.columns:
            api_params['columns'] = params.columns

        async with aiohttp.ClientSession() as session:
            for key, value in api_params.items():
                if isinstance(value, bool):
                    api_params[key] = str(value).lower()
                if isinstance(value, datetime.date):
                    api_params[key] = value.strftime('%Y-%m-%d')
            url_with_params = url + '?' + '&'.join([f'{k}={v}' for k, v in api_params.items()])
            # logger.debug(f"API call (Options Quotes): {url_with_params}")
            async with session.get(url, headers=self.headers, params=api_params) as response:
                self.api_calls += 1
                data, status_code = await self.handle_response_async(response, params.output)
                if not 200 <= status_code < 300:
                    logger.error(f"Error fetching options quotes for {params.option_symbol}. Status code: {status_code}")
                
        return data


    # /v1/stocks/candles/{resolution}/{symbol}/
    async def get_stock_candles(
        self,
        symbol: str,
        resolution: str = "1D",
        basic_params: Optional[BasicParams] = None,
        from_to_params: Optional[FromToParams] = None,
        columns: Optional[str] = None,
        exchange: Optional[str] = None,
        extended_hours: bool = False,
        exchange_country: str = "US",
        adjust_splits: Optional[bool] = None,
        adjust_dividends: Optional[bool] = None,
        output: str = "dataframe"
    ):
        """Get historical stock candles for a symbol asynchronously."""
        
        url = BASE_URL + f'stocks/candles/{resolution}/{symbol}/'
        params = {}
        if basic_params:
            params.update(basic_params.params)
        if from_to_params:
            params.update(from_to_params.params)
        
        options = {
            'columns': columns,
            'exchange': exchange,
            'extended_hours': extended_hours,
            'exchange_country': exchange_country,
            'adjust_splits': adjust_splits,
            'adjust_dividends': adjust_dividends
        }
    
        # Update params with non-None options
        for key, value in options.items():
            if value is not None:
                params[key] = value

        async with aiohttp.ClientSession() as session:
            for key, value in params.items():
                # Convert boolean values to lowercase strings
                if isinstance(value, bool):
                    value = str(value).lower()
                    params[key] = value
                    
            async with session.get(url, headers=self.headers, params=params) as response:
                # Construct full url with params for logging
                url_with_token = url + '&' + self.api_key
                logger.debug(f"API call (Candles): {url_with_token}")
                self.api_calls += 1
                return await self.handle_response_async(response, output)
            
    def get_stock_candles_parallel(self, symbols: List[str], resolution: str, from_date: datetime.date, to_date: datetime.date, max_concurrent: int = 50) -> Dict[str, pd.DataFrame | None]:
            async def fetch_stock(session, semaphore, symbol) -> Tuple[str, Any, int]:
                async with semaphore:
                    try:
                        data, status = await self.get_stock_candles(
                            symbol,
                            resolution=resolution,
                            basic_params=BasicParams(dateformat="timestamp", format="json"),
                            from_to_params=FromToParams(from_date=from_date, to_date=to_date),
                        )
                        return symbol, data, status
                    except Exception as e:
                        return symbol, None, str(e)

            async def get_price_histories_async() -> Dict[str, Any]:
                connector = TCPConnector(limit=max_concurrent)
                async with ClientSession(connector=connector) as session:
                    semaphore = asyncio.Semaphore(max_concurrent)
                    tasks = [fetch_stock(session, semaphore, symbol) for symbol in symbols]
                    return await asyncio.gather(*tasks)

            if sys.platform.startswith('win'):
                asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

            start_time = time()
            results = asyncio.run(get_price_histories_async())
            end_time = time()
            execution_time = end_time - start_time

            output = {}
            for symbol, data, status in results:
                if data is not None:
                    output[symbol] = data
                    logger.debug(f"{symbol}: Successfully fetched candles. Status: {status}")
                else:
                    output[symbol] = None
                    logger.warning(f"{symbol}: Failed to fetch candles. Status: {status}")

            logger.debug(f"Fetched candles for {len(symbols)} symbols in {execution_time:.2f} seconds")
            return output   
    
    def get_options_chains_parallel(self, params_list: List[OptionsChainParams], max_concurrent: int = 50) -> List[dict]:
        async def fetch_options_chain(session, params: OptionsChainParams):
            result = await self.get_options_chain(params)
            return params.id, params.underlying, result

        async def get_all_options_chains():
            connector = TCPConnector(limit=max_concurrent)
            async with ClientSession(connector=connector) as session:
                semaphore = asyncio.Semaphore(max_concurrent)
                async def bounded_fetch(params):
                    async with semaphore:
                        return await fetch_options_chain(session, params)
                tasks = [bounded_fetch(params) for params in params_list]
                return await asyncio.gather(*tasks)

        if sys.platform.startswith('win'):
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

        results = asyncio.run(get_all_options_chains())
        
        output = []
        for r in results:            
            _, _, data = r
            if data is not None:
                output.append(data)
        return output
    
    def get_options_quotes_parallel(self, params_list: List[OptionsQuoteParams], max_concurrent: int = 50) -> Dict[str, Any]:
        """Get options quotes for multiple option symbols asynchronously.

        Args:
            params_list (List[OptionsQuoteParams]): List of OptionsQuoteParams objects.
            max_concurrent (int, optional): Maximum number of concurrent requests. Defaults to 50.

        Returns:
            Dict[str, Any]: Dictionary containing the option symbol as the key and "data" and "status code" in a dict as the value.
        """
        async def fetch_options_quote(semaphore, params: OptionsQuoteParams):
            async with semaphore:
                result = await self.get_options_quotes(params)
                return params.option_symbol, result

        async def get_all_options_quotes():
            semaphore = asyncio.Semaphore(max_concurrent)
            tasks = [fetch_options_quote(semaphore, params) for params in params_list]
            return await asyncio.gather(*tasks)

        if sys.platform.startswith('win'):
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

        results = asyncio.run(get_all_options_quotes())

        output = {}
        for option_symbol, data in results:
            output[option_symbol] = data
        return output



import os
import re
from datetime import date, timedelta, datetime
import json
from typing import Dict, List

from loguru import logger
import pandas as pd

from marketdata.client import MarketDataClient
from marketdata.client_async import MarketDataAsyncClient
from marketdata.client_params import OptionsQuoteParams, OptionsChainParams
       
class MarketDataManager:
    
    def __init__(self, cache_dir: str="./data/cache"):
        self.client = MarketDataClient()
        self.client_async = MarketDataAsyncClient()
        self.candle_cache = {}
        self.first_available_cache = {}
        
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        if not os.path.exists(f"{cache_dir}/candles"):
            os.makedirs(f"{cache_dir}/candles")
    
    def validate_resolution(self, resolution: str):
        # Validate the input resolution
        pattern = re.compile(r"(\d+[MHWDY]?\b)")
        if not pattern.match(resolution):
            logger.error("Invalid resolution format. Must be in the format <number>[MHDWY]")
            return False
        return True
    
    def get_first_available_date(self, symbol: str, resolution: str):
        # Get the first available date for the symbol
        # This is useful for determining the earliest date that we can get data for
        # for a given symbol
        
        if not self.validate_resolution(resolution):
            logger.error("Invalid resolution format. Must be in the format <number>[MHDWY]")
            raise ValueError("Invalid resolution format. Must be in the format <number>[MHDWY]")
        
        if symbol not in self.first_available_cache:
        
            if not os.path.exists(f"./data/cache/candles/_first_available_{symbol}.json"):
                return None
            
            with open(f"./data/cache/candles/_first_available_{symbol}.json", "rb") as f:
                first_available = json.load(f)
                self.first_available_cache[symbol] = first_available
        
        first_available_date = self.first_available_cache[symbol].get(resolution, None)
        if first_available_date:
            return date.fromisoformat(first_available_date)
            
    def set_first_available_date(self, symbol: str, resolution: str, date: date):
        # Set the first available date for the symbol
        # This is useful for determining the earliest date that we can get data for
        # for a given symbol
        
        if not self.validate_resolution(resolution):
            logger.error("Invalid resolution format. Must be in the format <number>[MHDWY]")
            raise ValueError("Invalid resolution format. Must be in the format <number>[MHDWY]")
        
        if symbol not in self.first_available_cache:
            self.first_available_cache[symbol] = {}
        
        self.first_available_cache[symbol][resolution] = date.isoformat()
        
        with open(f"./data/cache/candles/_first_available_{symbol}.json", "w", encoding="utf-8") as f:
            # Serialize the date object to a string
            out_str = json.dumps(self.first_available_cache[symbol])
            # Write the string to the file
            f.write(out_str)
            
    
    def get_stock_candles(
        self,
        symbols: List[str],
        resolution: str,
        from_date: date | datetime,
        to_date: date | datetime | None = None,
        friendly_names=True,
        use_cache=True
    ) -> Dict[str, tuple[pd.DataFrame, int]]:
        if isinstance(from_date, datetime):
            from_date = from_date.date()
        if isinstance(to_date, datetime):
            to_date = to_date.date()
        
        if to_date is None:
            to_date = date.today()
            
        if from_date > to_date:
            logger.error("from_date cannot be after to_date")
            raise ValueError("from_date cannot be after to_date")
            # return {symbol: None for symbol in symbols}        
        
        if to_date > date.today():
            to_date = date.today()
        
        # Adjust for weekends and Mondays before 6pm
        if to_date.weekday() == 5:
            to_date -= timedelta(days=1)
        elif to_date.weekday() == 6:
            to_date -= timedelta(days=2)
        elif to_date.weekday() == 0 and datetime.now().time() < datetime.strptime("18:00", "%H:%M").time():
            to_date -= timedelta(days=3)
        
        self.validate_resolution(resolution)
        
        results = {}
        symbols_to_fetch = []
        
        for symbol in symbols:
            first_available_date = self.get_first_available_date(symbol, resolution)
            if first_available_date and from_date < first_available_date:
                logger.warning(f"Adjusting request date for {symbol} to available data from {first_available_date}")
                symbol_from_date = first_available_date
            else:
                symbol_from_date = from_date
            
            # Check in-memory cache
            cached_data = self.candle_cache.get(f"{symbol}_{resolution}", None)
            if cached_data:
                cached_from_date = cached_data["from_date"]
                cached_to_date = cached_data["to_date"]
                if isinstance(cached_from_date, str):
                    cached_from_date = date.fromisoformat(cached_from_date)
                if isinstance(cached_to_date, str):
                    cached_to_date = date.fromisoformat(cached_to_date)
                if symbol_from_date >= cached_from_date and to_date <= cached_to_date:
                    results[symbol] = cached_data["data"]
                                
            # Check file cache if not in memory
            if symbol not in results:
                file_list = [f for f in os.listdir("./data/cache/candles") if f.startswith(f"{symbol}_{resolution}")]
                for f in file_list:
                    parts = f.split("_")
                    from_date_file, to_date_file = map(date.fromisoformat, [parts[2], parts[3].split(".")[0]])
                    if symbol_from_date >= from_date_file and to_date <= to_date_file:
                        df = pd.read_parquet(f"./data/cache/candles/{f}")
                        self.candle_cache[f"{symbol}_{resolution}"] = \
                        {
                            "data": df,
                            "from_date": symbol_from_date,
                            "to_date": to_date_file
                        }
                        results[symbol] = df
                        break
            
            # If not found in cache, add to list to update via API request
            if symbol not in results:
                symbols_to_fetch.append(symbol)
        
        # Update symbols that were not found in the available cache
        if symbols_to_fetch:            
            self._update_candle_cache(symbols_to_fetch, resolution)
            for symbol in symbols_to_fetch:
                if symbol not in results:
                    results[symbol] = None
                else:
                    results[symbol] = self.candle_cache[f"{symbol}_{resolution}"]["data"]
        
        # Process each dataframe
        for symbol, df in results.items():
            if df is not None:
                
                if first_available_date is None:                    
                    try:
                        actual_from_date = df.at[0, 't']
                        self.set_first_available_date(symbol, resolution, date.fromisoformat(actual_from_date))
                    except AttributeError as e:
                        logger.error(f"Failed to set first available date for {symbol}")
                        logger.error(df)
                        raise e
                    
                if self.get_first_available_date(symbol, resolution) > from_date:
                    logger.warning(f"Requested data is not available for {symbol} from {from_date}. Using first available data from {self.get_first_available_date(symbol, resolution)} instead.")
                
                try:                    
                    if 't' in df.columns:                        
                        df.set_index('t', inplace=True)
                except AttributeError as e:
                    logger.error(f"Failed to set index for {symbol}")
                    raise e
                except KeyError as e:
                    print(results['SPY'])
                    raise e
                
                # Filter the data to the requested date range
                df = df[(df.index >= from_date.isoformat()) & (df.index <= to_date.isoformat())]
                df.reset_index(inplace=True)
                
                if friendly_names:
                    df = df.rename(columns={'t': 'datetime', 'o': 'open', 'h': 'high', 'l': 'low', 'c': 'close', 'v': 'volume'})
                
                results[symbol] = df
        
        return results
    
    def _update_candle_cache(self, symbols: List[str], resolution: str):
        
        logger.debug(f"Updating cache for {symbols} with resolution {resolution}")
        
        # Since we are updating the cache we want to fetch all available data
        from_date = date(2000, 1, 1)
        to_date = date.today()        
    
        fetched_data = self.client_async.get_stock_candles_parallel(symbols, resolution, from_date, to_date)
        for symbol, (data, status_code) in fetched_data.items():
            if data.get("s") == "no_data":
                logger.warning(f"No data available for {symbol} for resolution {resolution} and date range {from_date} to {to_date}")
                continue
            elif data.get("s") == "error":
                logger.error(f"Error fetching data for {symbol} for resolution {resolution} and date range {from_date} to {to_date}")
                logger.error(data)
                continue
            elif 200 <= status_code < 300:
                logger.info(f"Candle cache updated for {symbol} for resolution {resolution} and date range {from_date} to {to_date}")
                df = pd.DataFrame(data)
                actual_from_date = df.at[0, 't']
                actual_to_date = df.at[df.index[-1], 't']
                self.set_first_available_date(symbol, resolution, date.fromisoformat(actual_from_date))
                
                # Delete old files for this symbol and resolution
                old_files = [f for f in os.listdir(f"./data/cache/candles") if f.startswith(f"{symbol}_{resolution}")]
                for f in old_files:
                    os.remove(f"./data/cache/candles/{f}")
                
                # Write the data to the file cache
                df.to_parquet(f"./data/cache/candles/{symbol}_{resolution}_{actual_from_date}_{actual_to_date}.parquet")
                
                # And put the data into the in-memory cache
                self.candle_cache[f"{symbol}_{resolution}"] = {
                    "from_date": actual_from_date,
                    "to_date": actual_to_date,
                    "data": df
                }
            else:
                logger.error(f"Unknown status code for {symbol} for resolution {resolution} and date range {from_date} to {to_date}")
                logger.error(data)
                logger.error(status_code)
                raise ValueError(f"Unknown status code for {symbol} for resolution {resolution} and date range {from_date} to {to_date}")
    
    
    def get_option_chains(self, params: List[OptionsChainParams]) -> List[dict]:
        """
        Fetch options chains for multiple symbols in parallel.

        Args:
            params (List[OptionsChainParams]): A list of OptionsChainParams objects 
                                            specifying the options chains to retrieve.

        Returns:
            Dict[str, pd.DataFrame]: A dictionary where each key is a UUID (string) 
            generated for the request, and each value is a pandas DataFrame 
            containing the options chain data for the corresponding request.
            The UUID can be used to match the returned data with the input parameters.
        """
        return self.client_async.get_options_chains_parallel(params)

    def get_options_quotes(self, params: List[OptionsQuoteParams]) -> Dict[str, pd.DataFrame]:
        """
        Fetch options quotes for multiple options in parallel.

        Args:
            params (List[OptionsQuoteParams]): A list of OptionsQuoteParams objects 
                                            specifying the options quotes to retrieve.

        Returns:
            Dict[str, pd.DataFrame]: A dictionary where each key is the option symbol (string) 
            for which quotes were requested, and each value is dictionary with "data" and "status_code"
            values. "data" is a pandas DataFrame containing the quote data for that option symbol, 
            and "status_code" is the HTTP status code for the request.
        """
        return self.client_async.get_options_quotes_parallel(params)
    
    def get_api_call_count(self):
        return self.client.api_calls + self.client_async.api_calls
    
    
                    
if __name__ == "__main__":
    # mdm = MarketDataManager(API_KEY)

    # df, status_code = mdm.get_stock_candles("AAPL", "1D", date(2020, 1, 1))
    # logger.debug(df)
    # logger.debug(f"Status Code: {status_code}")
    # assert 200 <= status_code < 300
    pass